@@ -58,7 +58,7 @@ public class Execute {
             StartCommand.class,
             StatusCommand.class,
             StopCommand.class};
-    private static final Map<String, Class<?>> COMMANDS = new TreeMap<String, Class<?>>();
+    private static final Map<String, Class<?>> COMMANDS = new TreeMap<>();
 
     static {
         for (Class<?> c : COMMAND_CLASSES) {
@@ -86,6 +86,11 @@ public class Execute {
     private static final String PROP_KARAF_OPTS = "karaf.opts";
 
     public static void main(String[] args) throws Exception {
+        Package p = Package.getPackage("org.apache.karaf.instance.main");
+        if (p != null && p.getImplementationVersion() != null) {
+            System.setProperty("karaf.version", p.getImplementationVersion());
+        }
+
         if (args.length == 0) {
             listCommands();
             exit(0);
@@ -136,7 +141,7 @@ public class Execute {
 
     static void execute(InstanceCommandSupport command, File storageFile, String[] args) throws Exception {
         DefaultActionPreparator dap = new DefaultActionPreparator();
-        List<Object> params = new ArrayList<Object>(Arrays.asList(args));
+        List<Object> params = new ArrayList<>(Arrays.asList(args));
         params.remove(0); // this is the actual command name
 
         if (!dap.prepare(command, null, params)) {
@@ -163,7 +168,7 @@ public class Execute {
         if (exitAllowed) {
             System.exit(rc);
         } else {
-            throw new RuntimeException("" + rc);
+            throw new RuntimeException(Integer.toString(rc));
         }
     }
 
