@@ -19,7 +19,6 @@ package org.apache.karaf.management;
 import org.apache.karaf.jaas.config.KeystoreManager;
 import org.apache.karaf.management.internal.MBeanInvocationHandler;
 
-import java.io.Closeable;
 import java.io.IOException;
 import java.lang.reflect.Proxy;
 import java.net.BindException;
@@ -29,11 +28,11 @@ import java.net.ServerSocket;
 import java.net.Socket;
 import java.net.SocketAddress;
 import java.net.SocketException;
-import java.net.SocketImplFactory;
 import java.nio.channels.ServerSocketChannel;
 import java.rmi.server.RMIClientSocketFactory;
 import java.rmi.server.RMIServerSocketFactory;
 import java.security.GeneralSecurityException;
+import java.security.Security;
 import java.util.Enumeration;
 import java.util.Map;
 
@@ -53,17 +52,22 @@ import javax.rmi.ssl.SslRMIClientSocketFactory;
 
 public class ConnectorServerFactory {
 
-    private enum AuthenticatorType {NONE, PASSWORD, CERTIFICATE};
+    private enum AuthenticatorType {NONE, PASSWORD, CERTIFICATE}
 
     private MBeanServer server;
     private KarafMBeanServerGuard guard;
     private String serviceUrl;
+    private boolean jmxmpEnabled;
+    private String jmxmpServiceUrl;
     private String rmiServerHost;
     private Map<String, Object> environment;
+    private Map<String, Object> jmxmpEnvironment;
     private ObjectName objectName;
+    private ObjectName jmxmpObjectName;
     private boolean threaded = false;
     private boolean daemon = false;
     private JMXConnectorServer connectorServer;
+    private JMXConnectorServer jmxmpConnectorServer;
 
     private long keyStoreAvailabilityTimeout = 5000;
     private AuthenticatorType authenticatorType = AuthenticatorType.PASSWORD;
@@ -71,6 +75,8 @@ public class ConnectorServerFactory {
     private KeystoreManager keystoreManager;
     private String algorithm;
     private String secureProtocol;
+    private String[] enabledProtocols;
+    private String[] enabledCipherSuites;
     private String keyStore;
     private String trustStore;
     private String keyAlias;
@@ -99,6 +105,22 @@ public class ConnectorServerFactory {
         this.serviceUrl = serviceUrl;
     }
 
+    public boolean isJmxmpEnabled() {
+        return this.jmxmpEnabled;
+    }
+
+    public void setJmxmpEnabled(boolean jmxmpEnabled) {
+        this.jmxmpEnabled = jmxmpEnabled;
+    }
+
+    public String getJmxmpServiceUrl() {
+        return jmxmpServiceUrl;
+    }
+
+    public void setJmxmpServiceUrl(String jmxmpServiceUrl) {
+        this.jmxmpServiceUrl = jmxmpServiceUrl;
+    }
+
     public String getRmiServerHost() {
         return this.rmiServerHost;
     }
@@ -115,6 +137,14 @@ public class ConnectorServerFactory {
         this.environment = environment;
     }
 
+    public Map<String, Object> getJmxmpEnvironment() {
+        return this.jmxmpEnvironment;
+    }
+
+    public void setJmxmpEnvironment(Map<String,Object> jmxmpEnvironment) {
+        this.jmxmpEnvironment = jmxmpEnvironment;
+    }
+
     public ObjectName getObjectName() {
         return objectName;
     }
@@ -123,6 +153,14 @@ public class ConnectorServerFactory {
         this.objectName = objectName;
     }
 
+    public ObjectName getJmxmpObjectName() {
+        return this.jmxmpObjectName;
+    }
+
+    public void setJmxmpObjectName(ObjectName jmxmpObjectName) {
+        this.jmxmpObjectName = jmxmpObjectName;
+    }
+
     public boolean isThreaded() {
         return threaded;
     }
@@ -260,37 +298,55 @@ public class ConnectorServerFactory {
             this.server.registerMBean(this.connectorServer, this.objectName);
         }
 
+        if (jmxmpEnabled) {
+        	Security.addProvider(new PlainSaslServer.SaslPlainProvider());
+            JMXServiceURL jmxmpUrl = new JMXServiceURL(this.jmxmpServiceUrl);
+            this.jmxmpConnectorServer = JMXConnectorServerFactory.newJMXConnectorServer(jmxmpUrl, this.jmxmpEnvironment, guardedServer);
+            if (this.jmxmpObjectName != null) {
+                this.server.registerMBean(this.jmxmpConnectorServer, this.jmxmpObjectName);
+            }
+        }
+
         try {
             if (this.threaded) {
-                Thread connectorThread = new Thread() {
-                    public void run() {
-                        try {
-                            Thread.currentThread().setContextClassLoader(ConnectorServerFactory.class.getClassLoader());
-                            connectorServer.start();
-                        } catch (IOException ex) {
-                            if (ex.getCause() instanceof BindException){
-                                // we want just the port message
-                                int endIndex = ex.getMessage().indexOf("nested exception is");
-                                // check to make sure we do not get an index out of range
-                                if (endIndex > ex.getMessage().length() || endIndex < 0){
-                                    endIndex = ex.getMessage().length();
-                                }
-                                throw new RuntimeException("\n" + ex.getMessage().substring(0, endIndex) +
-                                                        "\nYou may have started two containers.  If you need to start a second container or the default ports are already in use " +
-                                                        "update the config file etc/org.apache.karaf.management.cfg and change the Registry Port and Server Port to unused ports");
-                            }
-                            throw new RuntimeException("Could not start JMX connector server", ex);
+                Thread connectorThread = new Thread(() -> {
+                    try {
+                        Thread.currentThread().setContextClassLoader(ConnectorServerFactory.class.getClassLoader());
+                        connectorServer.start();
+                        if (jmxmpEnabled && jmxmpConnectorServer != null) {
+                            jmxmpConnectorServer.start();
                         }
+                    } catch (IOException ex) {
+                        if (ex.getCause() instanceof BindException){
+                            // we want just the port message
+                            int endIndex = ex.getMessage().indexOf("nested exception is");
+                            // check to make sure we do not get an index out of range
+                            if (endIndex > ex.getMessage().length() || endIndex < 0){
+                                endIndex = ex.getMessage().length();
+                            }
+                            throw new RuntimeException("\n" + ex.getMessage().substring(0, endIndex) +
+                                                    "\nYou may have started two containers.  If you need to start a second container or the default ports are already in use " +
+                                                    "update the config file etc/org.apache.karaf.management.cfg and change the Registry Port and Server Port to unused ports");
+                        }
+                        throw new RuntimeException("Could not start JMX connector server", ex);
                     }
-                };
+                });
                 connectorThread.setName("JMX Connector Thread [" + this.serviceUrl + "]");
                 connectorThread.setDaemon(this.daemon);
                 connectorThread.start();
             } else {
                 this.connectorServer.start();
+                if (jmxmpEnabled && jmxmpConnectorServer != null) {
+                    jmxmpConnectorServer.start();
+                }
             }
         } catch (Exception ex) {
-            doUnregister(this.objectName);
+            if (this.objectName != null) {
+                doUnregister(this.objectName);
+            }
+            if (jmxmpEnabled && this.jmxmpObjectName != null) {
+                doUnregister(this.jmxmpObjectName);
+            }
             throw ex;
         }
     }
@@ -300,8 +356,16 @@ public class ConnectorServerFactory {
             if (this.connectorServer != null) {
                 this.connectorServer.stop();
             }
+            if (this.jmxmpEnabled && this.jmxmpConnectorServer != null) {
+                this.jmxmpConnectorServer.stop();
+            }
         } finally {
-            doUnregister(this.objectName);
+            if (this.objectName != null) {
+                doUnregister(this.objectName);
+            }
+            if (this.jmxmpEnabled && this.jmxmpObjectName != null) {
+                doUnregister(this.jmxmpObjectName);
+            }
         }
     }
 
@@ -310,6 +374,9 @@ public class ConnectorServerFactory {
             if (this.objectName != null && this.server.isRegistered(objectName)) {
                 this.server.unregisterMBean(objectName);
             }
+            if (this.jmxmpObjectName != null && this.server.isRegistered(jmxmpObjectName)) {
+                this.server.unregisterMBean(jmxmpObjectName);
+            }
         } catch (JMException ex) {
             // Ignore
         }
@@ -317,7 +384,8 @@ public class ConnectorServerFactory {
 
     private void setupSsl() throws GeneralSecurityException {
         SSLServerSocketFactory sssf = keystoreManager.createSSLServerFactory(null, secureProtocol, algorithm, keyStore, keyAlias, trustStore,keyStoreAvailabilityTimeout);
-        RMIServerSocketFactory rssf = new KarafSslRMIServerSocketFactory(sssf, isClientAuth(), getRmiServerHost());
+        RMIServerSocketFactory rssf = new KarafSslRMIServerSocketFactory(sssf, isClientAuth(), getRmiServerHost(), getEnabledProtocols(),
+                                                                         getEnabledCipherSuites());
         RMIClientSocketFactory rcsf = new SslRMIClientSocketFactory();
         environment.put(RMIConnectorServer.RMI_SERVER_SOCKET_FACTORY_ATTRIBUTE, rssf);
         environment.put(RMIConnectorServer.RMI_CLIENT_SOCKET_FACTORY_ATTRIBUTE, rcsf);
@@ -334,22 +402,40 @@ public class ConnectorServerFactory {
         private SSLServerSocketFactory sssf;
         private boolean clientAuth;
         private String rmiServerHost;
+        private String[] enabledProtocols;
+        private String[] enabledCipherSuites;
 
-        public KarafSslRMIServerSocketFactory(SSLServerSocketFactory sssf, boolean clientAuth, String rmiServerHost) {
+        public KarafSslRMIServerSocketFactory(SSLServerSocketFactory sssf, boolean clientAuth, String rmiServerHost, 
+                                              String[] enabledProtocols,
+                                              String[] enabledCipherSuites) {
             this.sssf = sssf;
             this.clientAuth = clientAuth;
             this.rmiServerHost = rmiServerHost;
+            this.enabledProtocols = enabledProtocols;
+            this.enabledCipherSuites = enabledCipherSuites;
         }
 
         public ServerSocket createServerSocket(int port) throws IOException {
             InetAddress host = InetAddress.getByName(rmiServerHost);
             if (host.isLoopbackAddress()) {
-                final SSLServerSocket ss = (SSLServerSocket) sssf.createServerSocket(port, 50);
+                final SSLServerSocket ss = (SSLServerSocket) sssf.createServerSocket(port, 50, host);
                 ss.setNeedClientAuth(clientAuth);
+                if (this.enabledProtocols != null && this.enabledProtocols.length > 0) {
+                    ss.setEnabledProtocols(this.enabledProtocols);
+                }
+                if (this.enabledCipherSuites != null && this.enabledCipherSuites.length > 0) {
+                    ss.setEnabledCipherSuites(this.enabledCipherSuites);
+                }
                 return new LocalOnlySSLServerSocket(ss);
             } else {
                 final SSLServerSocket ss = (SSLServerSocket) sssf.createServerSocket(port, 50, InetAddress.getByName(rmiServerHost));
                 ss.setNeedClientAuth(clientAuth);
+                if (this.enabledProtocols != null && this.enabledProtocols.length > 0) {
+                    ss.setEnabledProtocols(this.enabledProtocols);
+                }
+                if (this.enabledCipherSuites != null && this.enabledCipherSuites.length > 0) {
+                    ss.setEnabledCipherSuites(this.enabledCipherSuites);
+                }
                 return ss;
             }
         }
@@ -365,7 +451,7 @@ public class ConnectorServerFactory {
         public ServerSocket createServerSocket(int port) throws IOException {
             InetAddress host = InetAddress.getByName(rmiServerHost);
             if (host.isLoopbackAddress()) {
-                final ServerSocket ss = ServerSocketFactory.getDefault().createServerSocket(port, 50);
+                final ServerSocket ss = ServerSocketFactory.getDefault().createServerSocket(port, 50, host);
                 return new LocalOnlyServerSocket(ss);
             } else {
                 final ServerSocket ss = ServerSocketFactory.getDefault().createServerSocket(port, 50, InetAddress.getByName(rmiServerHost));
@@ -666,4 +752,20 @@ public class ConnectorServerFactory {
         throw new IOException("Only connections from clients running on the host where the RMI remote objects have been exported are accepted.");
     }
 
+    public String[] getEnabledProtocols() {
+        return enabledProtocols;
+    }
+
+    public void setEnabledProtocols(String[] enabledProtocols) {
+        this.enabledProtocols = enabledProtocols;
+    }
+
+    public String[] getEnabledCipherSuites() {
+        return enabledCipherSuites;
+    }
+
+    public void setEnabledCipherSuites(String[] enabledCipherSuites) {
+        this.enabledCipherSuites = enabledCipherSuites;
+    }
+    
 }
