@@ -74,13 +74,7 @@ import org.apache.karaf.features.internal.download.DownloadManager;
 import org.apache.karaf.features.internal.download.Downloader;
 import org.apache.karaf.features.internal.download.StreamProvider;
 import org.apache.karaf.features.internal.download.impl.DownloadManagerHelper;
-import org.apache.karaf.features.internal.model.Bundle;
-import org.apache.karaf.features.internal.model.Conditional;
-import org.apache.karaf.features.internal.model.ConfigFile;
-import org.apache.karaf.features.internal.model.Dependency;
-import org.apache.karaf.features.internal.model.Feature;
-import org.apache.karaf.features.internal.model.Features;
-import org.apache.karaf.features.internal.model.JaxbUtil;
+import org.apache.karaf.features.internal.model.*;
 import org.apache.karaf.features.internal.model.processing.FeaturesProcessing;
 import org.apache.karaf.features.internal.service.Blacklist;
 import org.apache.karaf.features.internal.service.Deployer;
@@ -112,6 +106,7 @@ import org.slf4j.LoggerFactory;
 import static java.util.Collections.singletonList;
 import static java.util.Comparator.comparing;
 import static java.util.jar.JarFile.MANIFEST_NAME;
+import static java.util.stream.Collectors.toMap;
 import static org.apache.karaf.profile.assembly.Builder.Stage.Startup;
 
 /**
@@ -189,19 +184,32 @@ public class Builder {
      * supported versions are defined.</p>
      */
     public enum JavaVersion {
-        Java16("1.6", 1),
-        Java17("1.7", 2),
-        Java18("1.8", 3),
-        Java9("9", 4),
-        Java10("10", 5),
-        Java11("11", 6);
+        Java6("1.6"),
+        Java7("1.7"),
+        Java8("1.8"),
+        Java9("9"),
+        Java10("10"),
+        Java11("11"),
+        Java12("12"),
+        Java13("13"),
+        Java14("14"),
+        Java15("15"),
+        Java16("16"),
+        Java17("17"),
+        Java18("18"),
+        Java19("19"),
+        Java20("20"),
+        Java21("21"),
+        Java22("22"),
+        Java23("23"),
+        Java24("24"),
+        Java25("25"),
+        Java26("26");
 
         private String version;
-        private int ordinal;
 
-        JavaVersion(String version, int ordinal) {
+        JavaVersion(String version) {
             this.version = version;
-            this.ordinal = ordinal;
         }
 
         public static JavaVersion from(String version) {
@@ -214,7 +222,7 @@ public class Builder {
         }
 
         public boolean supportsEndorsedAndExtLibraries() {
-            return this.ordinal < Java9.ordinal;
+            return compareTo(Java9) < 0;
         }
     }
 
@@ -286,6 +294,7 @@ public class Builder {
     Map<String, Stage> profiles = new LinkedHashMap<>();
     Map<String, RepositoryInfo> repositories = new LinkedHashMap<>();
     Map<String, Stage> features = new LinkedHashMap<>();
+    Set<String> firstStageBootFeatures = new HashSet<>();
     Map<String, Stage> bundles = new LinkedHashMap<>();
     List<String> blacklistedProfileNames = new ArrayList<>();
     List<String> blacklistedFeatureIdentifiers = new ArrayList<>();
@@ -293,7 +302,7 @@ public class Builder {
     List<String> blacklistedRepositoryURIs = new ArrayList<>();
     BlacklistPolicy blacklistPolicy = BlacklistPolicy.Discard;
     List<String> libraries = new ArrayList<>();
-    JavaVersion javase = JavaVersion.Java18;
+    JavaVersion javase = JavaVersion.Java8;
     KarafVersion karafVersion = KarafVersion.v4x;
     String environment = null;
     boolean useReferenceUrls;
@@ -311,6 +320,8 @@ public class Builder {
     String generateConsistencyReport;
     String consistencyReportProjectName;
     String consistencyReportProjectVersion;
+    // KARAF-7074: for recursive/inner features, we should use parallelism by default
+    int resolverParallelism = Math.max(2, Runtime.getRuntime().availableProcessors());
 
     private ScheduledExecutorService executor;
     private DownloadManager manager;
@@ -437,6 +448,17 @@ public class Builder {
         return this;
     }
 
+    /**
+     * Configure first stage features to use at boot stage. Each feature may be specified as
+     * <code>name</code> or <code>name/version</code> (no version ranges allowed).
+     * @param features
+     * @return
+     */
+    public Builder firstStageBootFeatures(String... features) {
+        this.firstStageBootFeatures.addAll(Arrays.asList(features));
+        return features(Stage.Boot, features);
+    }
+
     /**
      * Configure features to use at current {@link #defaultStage stage}. Each feature may be specified as
      * <code>name</code> or <code>name/version</code> (no version ranges allowed).
@@ -564,6 +586,11 @@ public class Builder {
         return this;
     }
 
+    public Builder resolverParallelism(final int resolverParallelism) {
+        this.resolverParallelism = resolverParallelism;
+        return this;
+    }
+
     /**
      * Configure builder to copy generated and configured profiles into <code>${karaf.etc}/profiles</code>
      * directory.
@@ -891,7 +918,7 @@ public class Builder {
         //
         MavenResolver resolver = createMavenResolver();
         manager = new CustomDownloadManager(resolver, executor, null, translatedUrls);
-        this.resolver = new ResolverImpl(new Slf4jResolverLog(LOGGER));
+        this.resolver = new ResolverImpl(new Slf4jResolverLog(LOGGER), resolverParallelism);
 
         //
         // Unzip KARs
@@ -944,6 +971,7 @@ public class Builder {
 
         boolean needFeaturesProcessorFileCopy = false;
         String existingProcessorDefinitionURI = null;
+        String additionalProcessorDefinitionURI = null;
         Path existingProcessorDefinition = etcDirectory.resolve("org.apache.karaf.features.xml");
         if (existingProcessorDefinition.toFile().isFile()) {
             existingProcessorDefinitionURI = existingProcessorDefinition.toFile().toURI().toString();
@@ -952,11 +980,12 @@ public class Builder {
         if (featuresProcessingLocation != null && featuresProcessingLocation.toFile().isFile()
                 && !featuresProcessingLocation.equals(existingProcessorDefinition)) {
             if (existingProcessorDefinitionURI != null) {
-                LOGGER.warn("Explicitly configured {} will be used for features processor configuration.", homeDirectory.relativize(featuresProcessingLocation));
+                LOGGER.info("Found additional features processor configuration: {}", homeDirectory.relativize(featuresProcessingLocation));
+                additionalProcessorDefinitionURI = featuresProcessingLocation.toFile().toURI().toString();
             } else {
                 LOGGER.info("Found features processor configuration: {}", homeDirectory.relativize(featuresProcessingLocation));
+                existingProcessorDefinitionURI = featuresProcessingLocation.toFile().toURI().toString();
             }
-            existingProcessorDefinitionURI = featuresProcessingLocation.toFile().toURI().toString();
             // when there are no other (configured via Maven for example) processing instructions (e.g., blacklisting)
             // we don't have to generate this file and may take original content
             needFeaturesProcessorFileCopy = true;
@@ -966,6 +995,30 @@ public class Builder {
         // configuration for bundle replacements or feature overrides.
         FeaturesProcessorImpl processor = new FeaturesProcessorImpl(existingProcessorDefinitionURI, null, blacklist, new HashSet<>());
 
+        if (additionalProcessorDefinitionURI != null) {
+            FeaturesProcessorImpl processor2 = new FeaturesProcessorImpl(additionalProcessorDefinitionURI, null, new Blacklist(), new HashSet<>());
+            FeaturesProcessing existingConfiguration = processor.getInstructions();
+            FeaturesProcessing additionalConfiguration = processor2.getInstructions();
+            additionalConfiguration.getBlacklistedRepositories()
+                    .forEach(br -> existingConfiguration.getBlacklistedRepositories().add(br));
+            additionalConfiguration.getBlacklistedRepositoryLocationPatterns()
+                    .forEach(lp -> existingConfiguration.getBlacklistedRepositoryLocationPatterns().add(lp));
+            additionalConfiguration.getBlacklistedFeatures()
+                    .forEach(bf -> existingConfiguration.getBlacklistedFeatures().add(bf));
+            additionalConfiguration.getBlacklistedBundles()
+                    .forEach(bb -> existingConfiguration.getBlacklistedBundles().add(bb));
+            additionalConfiguration.getFeatureReplacements().getReplacements()
+                    .forEach(fr -> existingConfiguration.getFeatureReplacements().getReplacements().add(fr));
+            additionalConfiguration.getBundleReplacements().getOverrideBundles()
+                    .forEach(br -> existingConfiguration.getBundleReplacements().getOverrideBundles().add(br));
+            additionalConfiguration.getOverrideBundleDependency().getRepositories()
+                    .forEach(r -> existingConfiguration.getOverrideBundleDependency().getRepositories().add(r));
+            additionalConfiguration.getOverrideBundleDependency().getFeatures()
+                    .forEach(f -> existingConfiguration.getOverrideBundleDependency().getFeatures().add(f));
+            additionalConfiguration.getOverrideBundleDependency().getBundles()
+                    .forEach(b -> existingConfiguration.getOverrideBundleDependency().getBundles().add(b));
+        }
+
         // add overrides from initialProfile
         Set<String> overrides = processOverrides(initialEffective.getOverrides());
         processor.addOverrides(overrides);
@@ -1227,14 +1280,12 @@ public class Builder {
                                         bundle2featureId.computeIfAbsent(bundle.getLocation().trim(), k -> new TreeSet<>()).add(feature.getId());
                                     }
                                 });
-                                feature.getConditional().forEach(cond -> {
-                                    cond.asFeature().getBundles().forEach(bundle -> {
-                                        // conditional bundles of feature
-                                        if (flavor.include(bundle)) {
-                                            bundle2featureId.computeIfAbsent(bundle.getLocation().trim(), k -> new TreeSet<>()).add(feature.getId());
-                                        }
-                                    });
-                                });
+                                feature.getConditional().forEach(cond -> cond.asFeature().getBundles().forEach(bundle -> {
+                                    // conditional bundles of feature
+                                    if (flavor.include(bundle)) {
+                                        bundle2featureId.computeIfAbsent(bundle.getLocation().trim(), k -> new TreeSet<>()).add(feature.getId());
+                                    }
+                                }));
                             }
                         });
                     }
@@ -1727,6 +1778,8 @@ public class Builder {
                     generatedDep.put(dep.getName(), dep);
                 }
                 dep.setDependency(false);
+                dep.setPrerequisite(firstStageBootFeatures.contains(dep.getName()) || firstStageBootFeatures.contains(
+                        nameOrPattern));
             }
         }
         // Add bundles
@@ -2055,7 +2108,12 @@ public class Builder {
                                 }
                             }
                             try (InputStream is = provider.open()) {
-                                Features featuresModel = JaxbUtil.unmarshal(url, is, false);
+                                Features featuresModel;
+                                if (JacksonUtil.isJson(url)) {
+                                    featuresModel = JacksonUtil.unmarshal(url);
+                                } else {
+                                    featuresModel = JaxbUtil.unmarshal(url, is, false);
+                                }
                                 // always process according to processor configuration
                                 featuresModel.setBlacklisted(processor.isRepositoryBlacklisted(url));
                                 processor.process(featuresModel);
@@ -2132,6 +2190,31 @@ public class Builder {
 
         // System bundle will be single bundle installed with bundleId == 0
         BundleRevision systemBundle = getSystemBundle();
+        if (resolverParallelism > 1) {
+            return doResolve(manager, resolver, repositories, features, bundles, optionals, processor, systemBundle);
+        }
+        // let a chance to be sequential in case order is important with the current framework
+        return features.stream()
+                .flatMap(it -> {
+                    try {
+                        return doResolve(manager, resolver, repositories, singletonList(it), bundles, optionals, processor, systemBundle).entrySet().stream();
+                    } catch (final RuntimeException e) {
+                        throw e;
+                    } catch (final Exception e) {
+                        throw new IllegalStateException(e);
+                    }
+                })
+                .collect(toMap(Map.Entry::getKey, Map.Entry::getValue, (a, b) -> a));
+    }
+
+    private Map<String, Integer> doResolve(DownloadManager manager,
+                                           Resolver resolver,
+                                           Collection<Features> repositories,
+                                           Collection<String> features,
+                                           Collection<String> bundles,
+                                           Collection<String> optionals,
+                                           FeaturesProcessor processor,
+                                           BundleRevision systemBundle) throws Exception {
         // Static distribution building callback and deployer that's used to deploy/collect startup-stage artifacts
         AssemblyDeployCallback callback = new AssemblyDeployCallback(manager, this, systemBundle, repositories, processor);
         Deployer deployer = new Deployer(manager, resolver, callback);
