@@ -16,16 +16,34 @@
  */
 package org.apache.karaf.shell.commands.impl;
 
-import org.apache.karaf.shell.commands.Command;
-import org.apache.karaf.shell.console.AbstractAction;
-import org.apache.karaf.shell.console.CloseShellException;
+import org.apache.karaf.shell.api.action.Action;
+import org.apache.karaf.shell.api.action.Command;
+import org.apache.karaf.shell.api.action.lifecycle.Reference;
+import org.apache.karaf.shell.api.action.lifecycle.Service;
+import org.apache.karaf.shell.api.console.Session;
+import org.apache.karaf.shell.support.ShellUtil;
+import org.slf4j.Logger;
+import org.slf4j.LoggerFactory;
 
 @Command(scope = "shell", name = "logout", description = "Disconnects shell from current session.")
-public class LogoutAction extends AbstractAction {
+@Service
+public class LogoutAction implements Action {
 
-    protected Object doExecute() throws Exception {
-        log.info("Disconnecting from current session...");
-        throw new CloseShellException();
+    private final Logger log = LoggerFactory.getLogger(getClass());
+
+    @Reference
+    Session session;
+
+    @Override
+    public Object execute() throws Exception {
+        boolean disableLogout = ShellUtil.loadPropertyFromShellCfg("disableLogout", Boolean::parseBoolean, false);
+        if (disableLogout) {
+            log.info("shell:logout disabled in org.apache.karaf.shell.cfg.");
+        } else {
+            log.info("Disconnecting from current session...");
+            session.close();
+        }
+
+        return null;
     }
-
 }
